/*	$Id$	*/

/*
 * Copyright (c) 2012, Internet Initiative Japan, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <string.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <unistd.h>

#include <libarms.h>
#include <axp_extern.h>
#include <transaction/transaction.h>
#include <protocol/arms_methods.h>

#include <module_db_mi.h>
#include <libarms_log.h>
#include <libarms/base64.h>
#include <libarms/malloc.h>

#include "compat.h"

module_cb_tbl_t mod_cb_tbl = {
	NULL,
	NULL
};

/* missing macro in sys/queue.h (montavista linux case.) */
#ifndef LIST_HEAD_INITIALIZER
#define LIST_HEAD_INITIALIZER(head) { NULL }
#endif
#ifndef LIST_FOREACH
#define	LIST_FOREACH(var, head, field)					\
	for ((var) = ((head)->lh_first);				\
		(var);							\
		(var) = ((var)->field.le_next))
#endif
#ifndef LIST_FIRST
#define	LIST_FIRST(head)		((head)->lh_first)
#endif

LIST_HEAD(module_storage_head, module_storage);

/*
 * current: module list.
 * addition: hopefully added modules.
 * new: updated module list
 * newconfig: updated module list for config.
 */
struct module_storage_head new =
	LIST_HEAD_INITIALIZER(module_storage_head);
struct module_storage_head newconfig =
	LIST_HEAD_INITIALIZER(module_storage_head);
struct module_storage_head addition =
	LIST_HEAD_INITIALIZER(module_storage_head);
struct module_storage_head current =
	LIST_HEAD_INITIALIZER(module_storage_head);

#define MAX_MODULE_INFO	8
#define CONFIG_ENCODE_BUF_SIZE	1365
#define CONFIGLEN_INIT_VAL	2147483647

struct arms_config_info {
	size_t len;
	size_t xml_len;
	char *buf;
	int flag;
	SIMPLEQ_ENTRY(arms_config_info) next;
};

struct module_storage {
	uint32_t id;
	char *ver;
	arms_module_info_t modinfo;
	int encoding;
	int permanent_running;
	int permanent_candidate;
	size_t config_candidate_xml_len;
	size_t config_total_xml_len;
	SIMPLEQ_HEAD(, arms_config_info) config;
	LIST_ENTRY(module_storage) chain;
};

static struct module_storage *
alloc_storage(void)
{
	struct module_storage *p;

	p = MALLOC(sizeof(*p));
	if (p == NULL)
		return NULL;

	p->permanent_candidate = ARMS_PERMANENT_INIT_VAL;
	p->config_candidate_xml_len = CONFIGLEN_INIT_VAL;

	memset(p, 0, sizeof(*p));
	SIMPLEQ_INIT(&p->config);
	return p;
}

static void
insert_to_current_by_id(struct module_storage *mod)
{
	struct module_storage *cur;
	struct module_storage *prev;

	if (mod == NULL)
		return;

	/* insert by ascending order */
	prev = NULL;
	LIST_FOREACH(cur, &current, chain) {
		if (mod->id < cur->id) {
			break;
		}
		prev = cur;
	}
	if (prev) {
		LIST_INSERT_AFTER(prev, mod, chain);
	} else {
		LIST_INSERT_HEAD(&current, mod, chain);
	}
}

static void
move_config(struct module_storage *dest, struct module_storage *src)
{
	struct arms_config_info *pconf;

	while ((pconf = SIMPLEQ_FIRST(&dest->config)) != NULL) {
		SIMPLEQ_REMOVE_HEAD(&dest->config, next);
		FREE(pconf);
	}
	dest->encoding = src->encoding;
	dest->permanent_candidate = src->permanent_candidate;
	dest->config_candidate_xml_len = src->config_candidate_xml_len;
	while ((pconf = SIMPLEQ_FIRST(&src->config)) != NULL) {
		SIMPLEQ_REMOVE_HEAD(&src->config, next);
		SIMPLEQ_INSERT_TAIL(&dest->config, pconf, next);
	}
}

static struct module_storage *
copy_storage(struct module_storage *src)
{
	struct module_storage *cpy;
	int i;

	cpy = alloc_storage();
	if (cpy == NULL)
		return NULL;

	cpy->id = src->id;
	if (src->ver != NULL)
		cpy->ver = STRDUP(src->ver);
	else
		cpy->ver = NULL;

	for (i = 0; i < src->modinfo.size; i++)
		cpy->modinfo.infostring[i] = STRDUP(src->modinfo.infostring[i]);
	cpy->modinfo.size = src->modinfo.size;

	return cpy;
}

static void
free_storage(struct module_storage *p)
{
	struct arms_config_info *pconf;
	int i;

	if (p == NULL)
		return;
	while ((pconf = SIMPLEQ_FIRST(&p->config)) != NULL) {
		SIMPLEQ_REMOVE_HEAD(&p->config, next);
		FREE(pconf);
	}
	if (p->ver)
		FREE(p->ver);
	for (i = 0; i < MAX_MODULE_INFO; i++)
		FREE((void *)p->modinfo.infostring[i]);
	FREE(p);

	return;
}

static void
free_storage_list(struct module_storage_head *h)
{
	struct module_storage *p;

	if (h == NULL)
		return;

	while ((p = LIST_FIRST(h)) != NULL) {
		LIST_REMOVE(p, chain);
		free_storage(p);
	}

	return;
}

static struct module_storage *
find_current(uint32_t id, const char*ver)
{
	struct module_storage *p;

	LIST_FOREACH(p, &current, chain) {
		if (p->id != id)
			continue;
		if (p->ver == NULL && ver != NULL)
			continue;
		if (p->ver != NULL && ver == NULL)
			continue;
		if (p->ver != NULL && strcmp(p->ver, ver) != 0)
			continue;

		return p;
	};

	return NULL;
}

static struct module_storage *
find_current_by_id(uint32_t id)
{
	struct module_storage *p;

	LIST_FOREACH(p, &current, chain) {
		if (p->id == id)
			break;
	};

	return p;
}

static int
reflect_module_config(int action, uint32_t mod_id, const char *mod_ver,
	    arms_module_info_t *modinfo, int encoding, const char *buf,
	    size_t len, int flag, arms_context_t *res,
	    char **catbuf, int *catlen, arms_base64_stream_t *base64)
{
	arms_config_cb_t func = res->callbacks.config_cb;
	int flag_tmp = flag;
	int err = 0;
	char binary_buff[AXP_BUFSIZE * 3 / 4 + 2 + 1]; /* + 2: modulo bytes */
	char *catbuf_tmp = NULL;

	if (!func) {
		return -1;
	}

	if (encoding == ARMS_DATA_BINARY) {
		memset(binary_buff, 0x00, sizeof(binary_buff));
		/* decode base64 */
		len = arms_base64_decode_stream(base64,
			binary_buff, sizeof(binary_buff) - 1,
			buf, len);
		if (len < 0) {
			libarms_log(ARMS_LOG_EBASE64_DECODE,
					"base64 decode error.");
			return -1;
		}
		binary_buff[len] = '\0';
		buf = binary_buff;
	}

	if (res->fragment == 0) {
		if ((len) || (*catbuf)) {
			catbuf_tmp = REALLOC(*catbuf, *catlen + len);
			if (catbuf_tmp == NULL) {
				/*Resource Exhausted*/
				return -1;
			}
		}
		*catbuf = catbuf_tmp;
		memcpy(*catbuf + *catlen, buf, len);
		*catlen += len;
		if (!(flag & AXP_PARSE_END)) {
			/* wait for next data */
			return 0;
		}
		flag |= ARMS_FRAG_FIRST;
		err = (*func)(mod_id,
				  mod_ver,		/* version */
				  modinfo,		/* module info */
				  action,
				  *catbuf, *catlen, flag, res->udata);

		if (err < 0) {
			return -1;
		}
	} else {
		do {
			int slen;
			/* call config callback */
			if (len > res->fragment) {
				slen = res->fragment;
				if (flag_tmp & AXP_PARSE_END) {
					flag_tmp &= ~AXP_PARSE_END;
				} else {
					flag_tmp |= ARMS_FRAG_CONTINUE;
				}
			} else {
				slen = len;
				/* if last fragment */
				if (flag & AXP_PARSE_END) {
					flag_tmp |= ARMS_FRAG_FINISHED;
				}
			}
			err = (*func)(mod_id,
					  mod_ver,		/* version */
					  modinfo,		/* module info */
					  action,
					  buf, slen, flag_tmp, res->udata);
			if (err < 0) {
				return -1;
			}
			buf += slen;
			len -= slen;
			flag &= ~ARMS_FRAG_FIRST;
			flag |= ARMS_FRAG_CONTINUE;
		} while (len > 0);
	}
	return 0;
}

int
add_module(int id, const char *ver, const char *url)
{
	struct module_storage *p, *pi;

	p = NULL;
	LIST_FOREACH(pi, &new, chain) {
		if (pi->id == id) {
			p = pi;
			break;
		}
	}

	if (p == NULL) {
		/* no module exists */
		p = alloc_storage();
		if (p == NULL)
			return -1;
		p->id = id;
		LIST_INSERT_HEAD(&new, p, chain);
	}

	if (ver != NULL)
		p->ver = STRDUP(ver);

	if (url != NULL && p->modinfo.size < MAX_MODULE_INFO)
		p->modinfo.infostring[p->modinfo.size++] = STRDUP(url);

	return 0;
}

int
add_permanent(int id, int permanent)
{
	struct module_storage *p;

	LIST_FOREACH(p, &newconfig, chain) {
		if (p->id == id) {
			break;
		}
	}
	if (p == NULL) {
		/* no module exists */
		p = alloc_storage();
		if (p == NULL)
			return -1;
		p->id = id;
		LIST_INSERT_HEAD(&newconfig, p, chain);
	}
	p->permanent_candidate = permanent;

	return 0;
}

int
add_config(int id, int encoding, const char *buf, size_t len, int flag)
{
	struct module_storage *p;
	struct arms_config_info *pconf;

	LIST_FOREACH(p, &newconfig, chain) {
		if (p->id == id) {
			break;
		}
	}
	if (p == NULL) {
		/* no module exists */
		p = alloc_storage();
		if (p == NULL)
			return -1;
		p->id = id;
		LIST_INSERT_HEAD(&newconfig, p, chain);
	}

	p->encoding = encoding;

	pconf = CALLOC(1, sizeof(*pconf) + len + sizeof('\0'));
	if (pconf == NULL) {
		return -1;
	}

	pconf->buf = ((char*)(pconf)) + sizeof(*pconf);
	pconf->flag = flag;
	if (p->encoding == ARMS_DATA_BINARY) {
		int i;
		for (i = 0; i < len; i++) {
			if (buf[i] != '\n') {
				pconf->buf[pconf->len] = buf[i];
				pconf->len++;
			}
		}
	} else {
		pconf->len = len;
		memcpy(pconf->buf, buf, len);
	}

	if (p->config_candidate_xml_len == CONFIGLEN_INIT_VAL) {
		p->config_candidate_xml_len = 0;
	}

	pconf->xml_len = strlen(arms_escape(buf));
	p->config_candidate_xml_len += pconf->xml_len;
	SIMPLEQ_INSERT_TAIL(&p->config, pconf, next);

	return 0;
}

static int
purge_module(uint32_t id)
{
	if (mod_cb_tbl.purge_module_cb != NULL) {
		return (*mod_cb_tbl.purge_module_cb)(id, mod_cb_tbl.udata);
	}

	return 0;
}

/*
 * update current module list.  sync with 'new'.
 */
int
sync_module(void)
{
	struct module_storage *p, *c, *n;
	int err;
	int failed = 0;

	/*
	 * first, lookup additional module and removal module.
	 */
	LIST_FOREACH(p, &new, chain) {
		c = find_current(p->id, p->ver);
		if (c) {
			move_config(p, c);
			p->permanent_running = c->permanent_running;
			p->config_total_xml_len = c->config_total_xml_len;

			/*
			 * skip use preloaded module
			 * (remove from current list)
			 */
			LIST_REMOVE(c, chain);
			free_storage(c);
			continue;
		}
		/*
		 * not find in current
		 * (add to addition list)
		 */
		n = copy_storage(p);
		if (n) {
			LIST_INSERT_HEAD(&addition, n, chain);
		}
	}

	/*
	 * now, modules in current list is NOT includes in new list.
	 * purge unused module (call user function)
	 */
	LIST_FOREACH(p, &current, chain) {
		if (arms_module_is_added(p->id))
			continue;
		err = purge_module(p->id);
		if (err) {
			failed = 1;
			continue;
		}
	}

	/* XXX: if (failed) ... */

	/*
	 * rebuild db
	 * copy new list to current list
	 */
	free_storage_list(&current);
	free_storage_list(&addition);
	while ((p = LIST_FIRST(&new)) != NULL) {
		LIST_REMOVE(p, chain);
		insert_to_current_by_id(p);
	}

	if (failed)
		return -1;
	return 0;
}

/*
 * update current config.  sync with 'newconfig'.
 */
void
sync_config(void)
{
	struct module_storage *p, *c;

	while ((p = LIST_FIRST(&newconfig)) != NULL) {
		c = find_current_by_id(p->id);
		if (c) {
			move_config(c, p);
		}
		LIST_REMOVE(p, chain);
		free_storage(p);
	}
}

void
update_running_module_info(void)
{
	struct module_storage *p;

	LIST_FOREACH(p, &current, chain) {
		if (p->config_candidate_xml_len != CONFIGLEN_INIT_VAL) {
			p->config_total_xml_len = p->config_candidate_xml_len;
			p->config_candidate_xml_len = CONFIGLEN_INIT_VAL;
		}
		if (p->permanent_candidate != ARMS_PERMANENT_INIT_VAL) {
			p->permanent_running = p->permanent_candidate;
			p->permanent_candidate = ARMS_PERMANENT_INIT_VAL;
		}
	}
}

static int
store_running_config(struct module_storage *p)
{
	arms_context_t *res = arms_get_context();
	int next;
	struct arms_config_info *pconf;
	int rv;
	char result[1023+1]; /* binary result size supposed to be (len % 3 == 0). */
	int blen;
	size_t total_len;

	next = ARMS_FRAG_FIRST;
	total_len = 0;
	p->config_candidate_xml_len = CONFIGLEN_INIT_VAL;

	do {
		memset(result, 0x00, sizeof(result));

		rv = res->callbacks.read_config_cb(
				p->id,
				ARMS_CONFIG_RUNNING,
				result,
				sizeof(result) - 1,
				&next,
				res->udata);

		if (ARMS_RESULT_IS_ERROR(rv)) {
			return -1;
		}

		if (ARMS_RESULT_IS_BYTES(rv) && (p->encoding == ARMS_DATA_BINARY)) {
			/* binary */
			blen = ARMS_RV_DATA_MASK(rv);
			pconf = CALLOC(1, sizeof(*pconf) + CONFIG_ENCODE_BUF_SIZE);
			if (pconf == NULL) {
				return -1;
			}
			pconf->buf = ((char*)(pconf)) + sizeof(*pconf);
			pconf->len = arms_base64_encode(pconf->buf, CONFIG_ENCODE_BUF_SIZE, result, blen);
		} else {
			/* text or error */
			pconf = CALLOC(1, sizeof(*pconf) + sizeof(result));
			if (pconf == NULL) {
				return -1;
			}
			pconf->buf = ((char*)(pconf)) + sizeof(*pconf);
			pconf->len = strlen(result);
			memcpy(pconf->buf, result, sizeof(result));
		}
		pconf->xml_len = strlen(arms_escape(result));
		SIMPLEQ_INSERT_TAIL(&p->config, pconf, next);
		total_len += pconf->xml_len;
		pconf->flag = next;

		if ((next & ARMS_FRAG_FINISHED)) {
			break;
		}

		next = ARMS_FRAG_CONTINUE;

	} while (1);

	p->config_total_xml_len = total_len;

	return 0;
}

int
complement_config(void)
{
	struct module_storage *p;
	struct arms_config_info *pconf;

	LIST_FOREACH(p, &current, chain) {
		if (!SIMPLEQ_EMPTY(&p->config)) {
			continue;
		}
		if (store_running_config(p) != 0) {
			while ((pconf = SIMPLEQ_FIRST(&p->config)) != NULL) {
				SIMPLEQ_REMOVE_HEAD(&p->config, next);
				FREE(pconf);
			}
			return -1;
		}
	}
	return 0;
}

int
remove_module_cb(uint32_t id, void *udata)
{
	arms_config_cb_t config_cb;
	arms_context_t *res = arms_get_context();
	int err = 0;

	res = udata;
	config_cb = res->callbacks.config_cb;
	if (config_cb) {
		err = (*config_cb) (id, "",
				    NULL,
				    ARMS_REMOVE_MODULE,
				    NULL, 0, 0,
				    res->udata);
	}
	return err;
}

int
purge_all_modules(void)
{
	struct module_storage *p;
	int err, failed;

	failed = 0;
	LIST_FOREACH(p, &current, chain) {
		err = purge_module(p->id);
		if (err) {
			failed = 1;
			continue;
		}
	}

	free_storage_list(&new);
	free_storage_list(&newconfig);
	free_storage_list(&addition);
	free_storage_list(&current);

	if (failed)
		return -1;
	return 0;
}

void
purge_all_config(void)
{
	struct module_storage *p;
	struct arms_config_info *pconf;

	LIST_FOREACH(p, &current, chain) {
		p->config_candidate_xml_len = CONFIGLEN_INIT_VAL;
		p->permanent_candidate = ARMS_PERMANENT_INIT_VAL;
		while ((pconf = SIMPLEQ_FIRST(&p->config)) != NULL) {
			SIMPLEQ_REMOVE_HEAD(&p->config, next);
			FREE(pconf);
		}
	}
	free_storage_list(&newconfig);
}

uint32_t
get_module_id(AXP *axp, int tag)
{
	const char *mod_idstr;
	uint32_t mod_id = 0;

	if (axp == NULL) {
		return 0;
	}

	/* module-id attribute */
	mod_idstr = axp_find_attr(axp, tag, "module-id");
	if (mod_idstr) {
		if (sscanf(mod_idstr, "%u", &mod_id) != 1) {
			sscanf(mod_idstr, "0x%x", &mod_id);
		}

		return mod_id;
	}

	mod_idstr = axp_find_attr(axp, tag, "id");
	if (mod_idstr) {
		if (sscanf(mod_idstr, "%u", &mod_id) != 1) {
			sscanf(mod_idstr, "0x%x", &mod_id);
		}
		return mod_id;
	}
	return 0;
}

uint32_t
get_module_order(AXP *axp, int tag)
{
	const char *mod_idstr;
	uint32_t mod_id = 0;

	if (axp == NULL) {
		return 0;
	}

	/* module-id attribute */
	mod_idstr = axp_find_attr(axp, tag, "commit-order");
	if (mod_idstr) {
		if (sscanf(mod_idstr, "%u", &mod_id) != 1) {
			sscanf(mod_idstr, "0x%x", &mod_id);
		}

		return mod_id;
	}
	return 0;
}

const char *
get_module_ver(AXP *axp, int tag)
{
	const char *ver_str;

	if (axp == NULL) {
		return 0;
	}

	/* module-id attribute */
	ver_str = axp_find_attr(axp, tag, "version");

	return ver_str;
}

int
get_permanent(AXP *axp, int tag)
{
	const char *str;
	int permanent = 0;

	if (axp == NULL) {
		return ARMS_PERMANENT_INVALID;
	}

	/* permanent attribute */
	str = axp_find_attr(axp, tag, "permanent");
	if (str) {
		if (sscanf(str, "%d", &permanent) != 1) {
			sscanf(str, "0x%x", &permanent);
		}
		return permanent;
	}
	return ARMS_PERMANENT_INVALID;
}

char *
lookup_module_ver(uint32_t id)
{
	struct module_storage *p;

	LIST_FOREACH(p, &current, chain) {
		if (p->id == id)
			return p->ver;
	}
	return NULL;
}

arms_module_info_t *
lookup_module_location(uint32_t id)
{
	struct module_storage *p;

	LIST_FOREACH(p, &current, chain) {
		if (p->id == id)
			return &p->modinfo;
	}
	return NULL;
}

int
init_module_cb(module_cb_tbl_t *tbl)
{
	if (tbl == NULL)
		return -1;

	memset(&mod_cb_tbl, 0, sizeof(mod_cb_tbl));
	mod_cb_tbl.purge_module_cb = tbl->purge_module_cb;
	mod_cb_tbl.udata = tbl->udata;

	return 0;
}

int
arms_count_module(void)
{
	struct module_storage *p;
	int i = 0;

	LIST_FOREACH(p, &current, chain) {
		i++;
	}
	return i;
}

/* ugly implementation */
int
arms_get_module_id(uint32_t *mod_id, int *permanent, int n)
{
	struct module_storage *p;

	LIST_FOREACH(p, &current, chain) {
		if (n-- > 0)
			continue;
		*mod_id = p->id;
		*permanent = p->permanent_running;
		return 0;
	}

	/* not found */
	return -1;
}

int
arms_get_permanent(void)
{
	struct module_storage *p;
	int permanent = -1;
	int mod_permanent;

	LIST_FOREACH(p, &current, chain) {
		if (p->permanent_candidate != ARMS_PERMANENT_INIT_VAL) {
			mod_permanent = p->permanent_candidate;
		} else {
			mod_permanent = p->permanent_running;
		}

		if (permanent == -1) {
			permanent = mod_permanent;
		} else if (permanent != mod_permanent) {
			return -1;
		}
	}
	return permanent;
}

size_t
arms_get_config_size(void)
{
	struct module_storage *p;
	size_t len = 0;

	LIST_FOREACH(p, &current, chain) {
		if (p->config_candidate_xml_len == CONFIGLEN_INIT_VAL) {
			len += p->config_total_xml_len;
		} else {
			len += p->config_candidate_xml_len;
		}
	}
	return len;
}

int
arms_reflect_config(arms_context_t *res, int action)
{
	struct module_storage *p;
	struct arms_config_info *pconf;
	int ret = -1;
	char *catbuf = NULL;
	int catlen = 0;
	arms_base64_stream_t base64;

	memset(&base64, 0x00, sizeof(base64));

	LIST_FOREACH(p, &current, chain) {
		SIMPLEQ_FOREACH(pconf, &p->config, next) {
			ret = reflect_module_config(action, p->id, p->ver,
					&p->modinfo, p->encoding, pconf->buf,
					pconf->len, pconf->flag, res,
					&catbuf, &catlen, &base64);
			if (ret != 0) {
				break;
			}
		}
		if (catbuf != NULL) {
			/* reset for next module id */
			FREE(catbuf);
			catbuf = NULL;
			catlen = 0;
			memset(&base64, 0x00, sizeof(base64));
		}
	}

	return ret;
}

size_t
arms_dump_xml_module(char *buf, int len)
{
	struct module_storage *p;
	size_t size, total;
	int i;
	const char *info;

	total = 0;
	LIST_FOREACH(p, &current, chain) {
		for (i = 0; i < p->modinfo.size; i++) {
			size = snprintf(buf, len,
			    "<module id=\"%d\" version=\"%s\">",
			    p->id, p->ver != NULL ? arms_escape(p->ver) : "");
			buf += size;
			len -= size;
			total += size;
			info = p->modinfo.infostring[i];
			size = snprintf(buf, len,
			    "%s</module>",
			    info != NULL ? arms_escape(info) : "");
			buf += size;
			len -= size;
			total += size;
		}
	}
	return total;
}

size_t
arms_dump_xml_config(char *buf, int len)
{
	struct module_storage *p;
	struct arms_config_info *pconf;
	size_t wrote = 0;

	LIST_FOREACH(p, &current, chain) {
		if (SIMPLEQ_EMPTY(&p->config)) {
			continue;
		}
		if (p->encoding == ARMS_DATA_BINARY) {
			/* binary */
			xmlprintf(buf, len, &wrote, NULL,
					"<md-config id=\"%d\" "
					"permanent=\"%d\" "
					"encoding=\"base64\">",
					p->id, p->permanent_running);
		} else {
			/* text or error */
			xmlprintf(buf, len, &wrote, NULL,
					"<md-config id=\"%d\" "
					"permanent=\"%d\">",
					p->id, p->permanent_running);
		}
		SIMPLEQ_FOREACH(pconf, &p->config, next) {
			wrote += strlcpy(buf + wrote, arms_escape(pconf->buf), len - wrote);
		}
		wrote += snprintf(buf + wrote, len - wrote, "</md-config>\n");
	}

	return wrote;
}

int
arms_module_is_added(int32_t id)
{
	struct module_storage *p;

	LIST_FOREACH(p, &new, chain) {
		if (p->id == id)
			return 1;
	}
	return 0;
}

int
arms_module_is_exist(int32_t id)
{
	struct module_storage *p;

	LIST_FOREACH(p, &current, chain) {
		if (p->id == id)
			return 1;
	}
	return 0;
}
